#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <math.h>

int i;

FILE* infile;
FILE* outfile;

unsigned char datablock[230408];
unsigned char outblock[153608];

char ChunkID[5],Format[5],Subchunk1ID[5],Subchunk2ID[5];
char SubChunkID[5];
char infilename[1024];
char outfilename[1024];

char inleafname[1024];
char outleafname[1024];

char settingsfilename[1024];
char appdirname[256];
char prefix[128];
char outdirname[256];
char leafname[128];

int ChunkSize,Subchunk1Size,AudioFormat;
int NumChannels,SampleRate,ByteRate,BlockAlign;
int BitsPerSample,BytesPerSample;

char BWAVEDescription[256];
char BWAVEOriginator[32];
char BWAVEOriginatorRef[32];
char BWAVEOriginationDate[10];
char BWAVEOriginationTime[8];
char UMID[64];
char BWAVEReserved[190];

int TimeRefLow, TimeRefHigh, BWAVEVersion,bhcount,chcount;
int BWAVEChunkSize,again,GeneralChunkSize,payload;
int again,bhcount;

int icut,icut2,slat,file_size;
int icut24, icut224;
int bwave_found,general_found,data_found,format_found,extended_found,header_size;
int blocksize,size_read,chunk_number;

int outsize16, sect_size, sect_number, number_of_sects;
int pairs_per_block, chunk_size,ok_to_process;
int carry_left, carry_right, nudge;
int erout_left, erin_left, sc_left, xnow_left;
int erout_right,erin_right,sc_right,xnow_right;
int tick,clipped,clip_amount;
int left_in,right_in,left_out,right_out;

unsigned int Subchunk2Size;

int read_header(void);
void read_settings(void);
int read_input_file_size(char*);
int scan_header(void);
int read_data_chunk(void);
int read_general_chunk(void);
int read_format_chunk(void);
int read_bwave_chunk(void);

void get4(char*);
int get4int(void);
int get3int(void);
int get2int(void);

void write4int(int);
void write2int(int);
void write4(char*);
void write_output_header(int);

void convert_block(void);
void save_block(void);

int clip_check(int);
int quantise16(int);
int dither8lsbits(void);

int main(int argc,char*argv[])
{
  icut=32768; icut2=2*icut;
  icut24=icut*256; icut224=icut2*256;
  erout_left=0; erout_right=0;
  erin_left=0; erin_right=0;
  tick=0;
  blocksize=115200;  pairs_per_block=blocksize/6;  /* N.B. Only processing 24bit */
  
  printf("WAVNoiseShaper v 0.10 (1st March 2017)\n");
  printf("Noise Shapes 24bit -> 16bit LPCM wave files\n");
  
  
  i=argc;
  printf("Number of arguments given = %d\n\n",i);

  if(i>0) printf("Application at  = %s\n",argv[0]);
  
  sprintf(outfilename,"%s",argv[0]);
  
  slat=(int)strrchr(outfilename,'/')-(int)outfilename;
  outfilename[slat]=(char)0;
  sprintf(appdirname,"%s",outfilename);
  sprintf(settingsfilename,"%s/Settings.txt",appdirname);
  
  printf("WAVNoiseShaper dir at %s\n",appdirname);
  printf("Settings file at %s\n",settingsfilename);
  
  read_settings();
  
  if(i>1)
  {
    printf("Input File at = %s\n",argv[1]);

    sprintf(infilename,"%s",argv[1]);
    
    sprintf(leafname,"%s",1+strrchr(infilename,'/'));
    printf("leafname = %s\n",leafname);
    
    sprintf(outfilename,"%s/%s_%s",outdirname,prefix,leafname);
    printf("Output to file\n  %s\n",outfilename);
    
    file_size=read_input_file_size(infilename);
    
    printf("Input file size = %d bytes\n",file_size);
    
    infile=fopen(infilename,"rd");
    
    scan_header();
    
    printf("Chunk types found [1 = yes : 0 = no]\n");
    printf("   format = %d",format_found);
    printf("   bwave  = %d",bwave_found);
    printf("   extended = %d",extended_found);
    printf("   others = %d",general_found);
    printf("   data   = %d \n",data_found);
    header_size=file_size-payload;
    printf("Header size = %d Bytes\n",header_size);
    
    ok_to_process=1;
    
    if(BitsPerSample!=24)
    {
      ok_to_process=0;
      printf("Payload = %d bits per sample: NOT suitable for Shaping!\n",BitsPerSample);
    }
    
    
    if(NumChannels!=2)
    {
      ok_to_process=0;
      printf("Not stereo, has %d channels: NOT suitable for Shaping!\n",NumChannels);
    }
    
    printf("Sample Rate = %d\n\n",SampleRate);
      
    if(ok_to_process==1)
    { 
      
      printf("Making Noise Shaped 16bit version\n");
      outfile=fopen(outfilename,"wd");
      write_output_header(payload);
      
     
      do
      {
        
        chunk_size=load_block();
        
        outsize16 = (2*chunk_size)/3;
        
        convert_block();
        
        save_block();
        
        chunk_number++;
        
        if(clipped==1)
        {
          printf(" * Clipped! %6.1lf by %d *\n",0.1*(double)chunk_number,clip_amount);
        }
        
        tick++;
        
        if(tick==50)
        {
          printf(" %6.1lf\n", 0.1*(double)chunk_number);
          tick=0;
        }
      
      } while (chunk_size==blocksize);
      
      fclose(outfile);
    }
    
    
    fclose(infile);
  
  }
  else
  {
    printf("No input file!\n");
    printf("Drag a WAV file to the WAVCleaner application icon\n");
    printf("to see the results\n");
  }
  
  
       
  printf("Done!\n");
  return 0;
}

void save_block(void)
{
  fwrite(outblock,sizeof(unsigned char),outsize16,outfile);
}


int load_block(void)
{
  /* This procedure loads the next 0.1 second block of
     data from the input file. */
  
  int block_size_read;
  block_size_read=fread(datablock,sizeof(unsigned char),blocksize,infile);
  return block_size_read;
}


void convert_block(void)
{
  /* This procedure converts the values in the data block
     into integers, ready for examination. */
  
  int io,ii_in,ii_out,lt,rt;
  unsigned char l1,l2,r1,r2;
  ii_in=0; ii_out=0; io=0; clipped=0; clip_amount=0;
  
  do
  {
    lt= 65536*(int)datablock[ii_in+2] + 256*(int)datablock[ii_in+1] + (int)datablock[ii_in];
    ii_in+=3;
    
   
    rt= 65536*(int)datablock[ii_in+2] +256*(int)datablock[ii_in+1] + (int)datablock[ii_in];
    ii_in+=3;
     
    if (lt>=icut24) lt-=icut224;
    if (rt>=icut24) rt-=icut224;
    left_in=lt;    right_in=rt;
    
    erout_left=erin_left; erout_right=erin_right;       /* update delays */
    
    sc_left=left_in-erout_left;
    left_out=quantise16( sc_left + dither8lsbits() );   /* shape left channel */
    erin_left= (256 * left_out) - sc_left;
    
    sc_right=right_in-erout_right;
    right_out=quantise16( sc_right + dither8lsbits() ); /* shape right channel */
    erin_right= (256 * right_out) - sc_right; 
    
    lt=left_out; rt=right_out;
    
    lt=clip_check(lt);
    rt=clip_check(rt);
        
    if(lt<0) lt+=2*icut;
    if(rt<0) rt+=2*icut;
    
    l1=(unsigned char)((lt>>8)&255);
    l2=(unsigned char)(lt&255);
    r1=(unsigned char)((rt>>8)&255);
    r2=(unsigned char)(rt&255);
    
    outblock[ii_out]=l2;
    outblock[ii_out+1]=l1;
    outblock[ii_out+2]=r2;
    outblock[ii_out+3]=r1;
    
    ii_out+=4;
    
    io++;
  } while (io<pairs_per_block);
    
    
}





int clip_check(int insample)
{
  int outsample, icd;
  
  icd=0; outsample=insample;
  
  if( insample > 32767)
  {
    outsample=32767;
    clipped=1;
    icd=insample-outsample;
  }
  
  if(icd > clip_amount) clip_amount=icd;
  
  if( insample < -32767)
  {
    outsample= -32767;
    clipped=1;
    icd=outsample-insample;
  }
  
  if(icd > clip_amount) clip_amount=icd;
    
  return outsample;
  
}


int quantise16(int iq)
{
  int answer;
  
  if (iq>=0)
  {
    iq+=128;
  }
  else
  {
    iq-=128;
  }
  
  answer=iq/256;
  
  return answer;
  
}

int dither8lsbits(void)
{
  /* This procedure returns a 'random dither'
  levet as an integer in the range +256 to -256 */
  
  int answer;
  
  answer=rand()%256 + rand()%256 - 256;
  return answer;
}




int scan_header(void)
{
  int headtype;
  headtype=0; bhcount=0; again=0;
  header_size=0;
  bwave_found=0; general_found=0; data_found=0; format_found=0; extended_found=0;

  get4(ChunkID);
  printf("\nChunkID = %s\n",ChunkID);
  ChunkSize=get4int();
  printf("Chunk Size = %d\n",ChunkSize);
  get4(Format);
  printf("Format = %s\n\n\n",Format);
  
  if (strcmp(Format,"WAVE")==0)
  {
    printf("WAVE found\n");
    
    do
    {
      
      get4(SubChunkID);
      printf("\n\nSubchunkID = %s\n",SubChunkID);
      again=0;
      
      if(strcmp(SubChunkID,"bext\0")==0)
      {
        printf("\nBWAVE Chunk found\n");
        bwave_found=1;
        again=1;
        read_bwave_chunk();
      }
      
      if(strcmp(SubChunkID,"fmt \0")==0)
      {
        printf("Format chunk found\n");
        format_found=1;
        again=1;
        read_format_chunk();
      }
      
      
      
      if(again==0)
      {
        if(strcmp(SubChunkID,"data\0")==0)
        {
          printf("Data chunk found\n");
          data_found=1;
          read_data_chunk();
        }
        else
        {
          printf("General chunk found\n");
          general_found=1;
          read_general_chunk();
          again=1;
        }
      }
       
      
    
    } while (again==1);
  }
  
  return headtype;
}


int read_data_chunk(void)
{
  Subchunk2Size=get4int();
  printf("Subchunk2Size = %d [Number of Data payload bytes]\n",Subchunk2Size);
  payload=Subchunk2Size; 
  return 1;
}


int read_general_chunk(void)
{
  char c1;
  GeneralChunkSize=get4int();
  printf("Size = %d\n",GeneralChunkSize);
  bhcount=0;
  do
  {
    fscanf(infile,"%c",&c1);
    if ((int)c1!=0) printf("%c",c1);
    bhcount++;
  } while (bhcount<GeneralChunkSize);
  printf("\n");
  return 1;
}


int read_format_chunk(void)
{
   char c1;
   int wxvalue;
    Subchunk1Size=get4int();
    printf(" Sub chunk size = %d\n",Subchunk1Size);
    bhcount=0;
    
    AudioFormat=get2int(); bhcount+=2;
    
    if(AudioFormat==1)
    {
      printf("Simple LPCM format OK\n");
    }
    else
    {
      printf("  *** Not Simple Wave file!  ***\n");
      
      if(AudioFormat==65534)
      {
        extended_found=1;
        printf("Extensible Wave format found\n");
        AudioFormat=1;
      }
      
      
    }
    
    NumChannels=get2int(); bhcount+=2;
    printf("Number of Channels = %d\n",NumChannels);
    SampleRate=get4int(); bhcount+=4;
    printf("Sample Rate = %d\n",SampleRate);
    ByteRate=get4int(); bhcount+=4;
    printf("Byte Rate = %d\n",ByteRate);
    
    BlockAlign=get2int(); bhcount+=2;
    printf("Block Align = %d\n",BlockAlign);
    BitsPerSample=get2int(); bhcount+=2;
    printf("Bits per sample = %d\n",BitsPerSample);
    BytesPerSample=BitsPerSample/8;
    printf("Bytes per sample = %d\n",BytesPerSample);
    
    if(extended_found==1)
    {
      wxvalue=get2int();
      printf("cbSize = %d\n",wxvalue);
      bhcount+=2;
      wxvalue=get2int();
      printf("wValidBitsPerSample = %d\n",wxvalue);
      bhcount+=2;
      wxvalue=get4int();
      printf("dwChannelMask = %d\n",wxvalue);
      bhcount+=4;
      
      
    }
    
    while(bhcount<Subchunk1Size)
    {
      bhcount++;
      fscanf(infile,"%c",&c1);
    }
    
    printf("bhcount %d\n",bhcount);
  
  return 1;
}

int read_bwave_chunk(void)
{
  char c1;
  printf("Reading Broadcast Wav header chunk\n");
    
    BWAVEChunkSize=get4int();
    
    printf("Size = %d\n",BWAVEChunkSize);
    
    bhcount=0;
   
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEDescription[i]=c1;
      i++; bhcount++;
    } while (i<256);
    
    printf("%3d Description => %s\n",bhcount,BWAVEDescription);
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEOriginator[i]=c1;
      i++; bhcount++;
    } while (i<32);
    
    printf("%3d Originator => %s\n",bhcount,BWAVEOriginator);
    
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEOriginatorRef[i]=c1;
      i++; bhcount++;
    } while (i<32);
    
    printf("%3d Reference => %s\n",bhcount,BWAVEOriginatorRef);
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEOriginationDate[i]=c1;
      i++; bhcount++;
    } while (i<10);
    
    printf("%3d Date => %s\n",bhcount,BWAVEOriginationDate);
    
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEOriginationTime[i]=c1;
      i++; bhcount++;
    } while (i<8);
    
    printf("%3d Time => %s\n",bhcount,BWAVEOriginationTime);
    
    TimeRefLow=get4int(); bhcount+=4;
    TimeRefHigh=get4int(); bhcount+=4;
    
    printf("%3d Time ref values  %d  %d\n",bhcount,TimeRefHigh,TimeRefLow);
    
    BWAVEVersion=get2int(); bhcount+=2;
    
    printf("%3d BWAVE Version = %d\n",bhcount,BWAVEVersion);
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      UMID[i]=c1;
      
      if((int)c1>31) printf("%d %c\n",bhcount,c1);
      
      i++; bhcount++;
    } while (i<64);
    
    printf("%3d Read UMID\n",bhcount);
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEReserved[i]=c1;
      
      if((int)c1>31) printf("%d %c\n",i,c1);
      
      i++; bhcount++;
    } while (i<190);
    
    printf("%3d Read BWAVE Reserved\n",bhcount);
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      if((int)c1!=0) printf("%d %d %c\n",bhcount,(int)c1,c1);
      i++; bhcount++;
    } while (bhcount<BWAVEChunkSize);
    
    printf("Reached end of BWAVE chunk header\n");
  
  return 1;
}

int read_input_file_size(char* filename)
{
  struct stat fileinfo;
  int result;
  result=-1;
  if (stat(filename,&fileinfo)==0)
  {
    result=(int)fileinfo.st_size;
    printf("file %s is\n   %d bytes int\n",filename,result);
  }
  else
  {
    printf("file %s not found!\n",filename);
  }
  
  return result;
}

void read_settings(void)
{
  infile=fopen(settingsfilename,"rt");
  
  fscanf(infile,"%s",prefix);
  printf("Output prefix = %s\n",prefix);
  fscanf(infile,"%s",outdirname);
  printf("Output to dir = %s\n",outdirname);
  
  
  fclose(infile);
}


int read_header(void)
{
  int irh;

   printf("Reading header\n");
   
   ByteRate=0;
    
  irh=ByteRate;
  return irh;
}


int get2int(void)
{
  int result;
  unsigned char cip1,cip2;
  
  fscanf(infile,"%c",&cip1);
  fscanf(infile,"%c",&cip2);

  result=(int)cip2;
  result=(int)cip1+256*result;

  return result;
} 

int get4int(void)
{
  int result;
  unsigned char cip1,cip2,cip3,cip4;
  
  fscanf(infile,"%c",&cip1);
  fscanf(infile,"%c",&cip2);
  fscanf(infile,"%c",&cip3);
  fscanf(infile,"%c",&cip4);
    
  result=(int)cip4;
  result=(int)cip3+256*result;
  result=(int)cip2+256*result;
  result=(int)cip1+256*result;
  
  return result;
} 
void get4(char* outstring)
{
  int isp;
  char csp;
  isp=0;
  do
  {
    fscanf(infile,"%c",&csp);
    outstring[isp]=csp;
    isp++;
  } while (isp<4);
  outstring[4]=(char)0;

}


void write4int(int inv)
{
  char cop1,cop2,cop3,cop4;
  
  cop1=(unsigned char)(inv&255);
  cop2=(unsigned char)((inv>>8)&255);
  cop3=(unsigned char)((inv>>16)&255);
  cop4=(unsigned char)((inv>>24)&255);
  
  fprintf(outfile,"%c",cop1);
  fprintf(outfile,"%c",cop2);
  fprintf(outfile,"%c",cop3);
  fprintf(outfile,"%c",cop4);
}


void write2int(int inv)
{
  unsigned char cop1,cop2;
  cop1=(unsigned char)(inv&255);
  cop2=(unsigned char)((inv>>8)&255);
  fprintf(outfile,"%c",cop1);
  fprintf(outfile,"%c",cop2);
}

void write4(char* outstring)
{
  int isp;
  
  isp=0;
  do
  {
    fprintf(outfile,"%c",outstring[isp]);
    isp++;
  } while (isp<4);
  
}


void write_output_header(int plo)
{
  
  printf("Payload to write = %d\n\n",plo);
  write4("RIFF\0");
  write4int(plo+36);
  write4("WAVE\0");
  write4("fmt \0");
  write4int(16);
  write2int(AudioFormat);
  write2int(NumChannels);
  write4int(SampleRate);
  
  write4int( ((2*ByteRate)/3) );
  printf("Output Byte rate = %d\n",(2*ByteRate)/3);
  
  write2int( (2*BlockAlign/3) );
  printf("Output Block Align = %d\n", (2*BlockAlign)/3);
  
  write2int( (2*BitsPerSample)/3) ;
  printf("Output bits per sample = %d\n", (2*BitsPerSample)/3);
  
  write4("data");
  
  write4int( (2*Subchunk2Size)/3);
  printf("Output payload = %d\n",(2*Subchunk2Size)/3);
  
}








